require 'fluent/plugin/output'
require 'fluent/plugin_helper/thread'
require 'net/http'
require 'json'

class SentryException
  attr_accessor :values

  def initialize(exception)
    self.values = []
    trace = {}
    trace['stacktrace'] = {}
    trace['stacktrace']['frames'] = []
    values.push(trace)

    frames = exception['frames']
    p exception
    p frames

    frames.each do |frame|
      add_values(frame)
    end

    add_error_type_and_message(exception['error_type'], exception['error_message'])
    values[0]['mechanism'] = {}
    values[0]['mechanism']['type'] = 'generic'
    values[0]['mechanism']['handled'] = true
  end

  def add_values(frame)
    sentry_frame = {
      in_app: frame['in_app'] || false,
      function: frame['function_name'] || nil,
      filename: frame['file_name'] || nil,
      lineno: frame['line_number'] || nil,
      colno: frame['column_number'] || nil
    }

    values[0]['stacktrace']['frames'].push(sentry_frame)
  end

  def add_error_type_and_message(type, message)
    values[0]['type'] = type
    values[0]['value'] = message
  end

  def to_json(*_args)
    hash = {}
    instance_variables.each do |var|
      hash[var.to_s.delete '@'] = instance_variable_get var
    end
    hash.to_json
  end
end

class SentryMessageFormat
  attr_accessor :event_id, :server_name, :timestamp, :environment,
                :release, :tags, :request, :user, :transaction, :exception,
                :error_type, :error_message, :platform

  def initialize(
    timestamp,
    event_id,
    component_name,
    environment,
    release,
    tags,
    request,
    exception,
    platform,
    transaction
  )
    @event_id = event_id
    @timestamp = timestamp
    @server_name = component_name
    @environment = environment
    @release = release
    @tags = tags
    @request = request
    @platform = platform
    @transaction = transaction
    self.exception = SentryException.new(exception)
  end

  def to_json(*_args)
    hash = {}
    instance_variables.each do |var|
      hash[var.to_s.delete '@'] = instance_variable_get var
    end

    hash.to_json
  end
end

module Fluent::Plugin
  class SentryOutput < Output
    Fluent::Plugin.register_output('sentry', self)

    helpers :thread

    config_param :ingest_url, :string
    config_param :auth_token, :string
    config_param :project_id, :string

    def write(chunk)
      chunk.each do |time, record|
        log.debug time, record
        begin
          send_to_sentry(record, time)
        rescue => e
          log.error 'Unable to send event to Sentry, Err class - ', e.class, ', error message - ', e.message, ', error backtrace - ', e.backtrace
        end
      end
    end

    def create_sentry_event_ingestion_url
      "#{@ingest_url}/api/#{@project_id}/store/?sentry_key=#{@auth_token}&sentry_version=7"
    end

    def send_to_sentry(record, time)
      event = SentryMessageFormat.new(
        record['timestamp'] || Time.at(time).to_i || Time.now.to_i,
        SecureRandom.hex(16),
        record['component_name'] || nil,
        record['environment'] || nil,
        record['commitId'] || nil,
        record['tags'] || nil,
        record['request'] || nil,
        record['exception'] || nil,
        record['platform'] || 'node',
        record['transaction'] || nil
      )

      url = create_sentry_event_ingestion_url
      request_payload = event.to_json

      log.debug 'Sending payload - ', request_payload, ' to - ', url
      Net::HTTP.post(URI(url), request_payload)
    end

  end
end
